@@ -11,47 +11,73 @@
 # FOR A PARTICULAR PURPOSE.
 #
 ##############################################################################
-from __future__ import print_function
 
 import doctest
+import glob
+import importlib.util
 import os
 import re
-import signal
 import shutil
+import signal
 import subprocess
 import sys
 import tempfile
 import unittest
 from contextlib import contextmanager
 
-import ZConfig
 import manuel.capture
 import manuel.doctest
 import manuel.testing
 import zc.customdoctests
-import zdaemon
+import ZConfig
 from zope.testing import renormalizing
 
-try:
-    import pkg_resources
-    zdaemon_loc = pkg_resources.working_set.find(
-        pkg_resources.Requirement.parse('zdaemon')).location
-    zconfig_loc = pkg_resources.working_set.find(
-        pkg_resources.Requirement.parse('ZConfig')).location
-except (ImportError, AttributeError):
-    zdaemon_loc = os.path.dirname(os.path.dirname(zdaemon.__file__))
-    zconfig_loc = os.path.dirname(os.path.dirname(ZConfig.__file__))
+import zdaemon
+
+
+_zdaemon_spec = importlib.util.find_spec('zdaemon')
+_zdaemon_origin = _zdaemon_spec.origin if _zdaemon_spec else zdaemon.__file__
+_zdaemon_package_dir = os.path.dirname(_zdaemon_origin)
+zdaemon_loc = os.path.dirname(_zdaemon_package_dir)
+
+_zconfig_spec = importlib.util.find_spec('ZConfig')
+_zconfig_origin = _zconfig_spec.origin if _zconfig_spec else ZConfig.__file__
+_zconfig_package_dir = os.path.dirname(_zconfig_origin)
+zconfig_loc = os.path.dirname(_zconfig_package_dir)
 
 
 def write(name, text):
     with open(name, 'w') as f:
         f.write(text)
 
+
 def read(name):
     with open(name) as f:
         return f.read()
 
 
+def test_ChildExits():
+    """
+    ``_ChildExits`` is used to record the exit status for a
+    foreign process when it was captured by the ``SIGCHLD``handler
+    due to a race condition.
+
+    >>> from zdaemon.zdrun import _ChildExits
+    >>> exits = _ChildExits()
+
+    Initially, we have no exit status for a process.
+    >>> exits.fetch(5)
+
+    After we have recorded exit information, ``fetch`` will
+    retrieve and reset it.
+    >>> exits[5] = 0
+    >>> exits.fetch(5)
+    0
+    >>> exits.fetch(5)
+
+    """
+
+
 def make_sure_non_daemon_mode_doesnt_hang_when_program_exits():
     """
     The whole awhile bit that waits for a program to start
@@ -69,6 +95,7 @@ def make_sure_non_daemon_mode_doesnt_hang_when_program_exits():
 
     """
 
+
 def dont_hang_when_program_doesnt_start():
     """
     If a program doesn't start, we don't want to wait for ever.
@@ -88,6 +115,7 @@ def dont_hang_when_program_doesnt_start():
 
     """
 
+
 def allow_duplicate_arguments():
     """
     Wrapper scripts will often embed configuration arguments. This could
@@ -112,6 +140,7 @@ def allow_duplicate_arguments():
 
     """
 
+
 def test_stop_timeout():
     r"""
 
@@ -152,45 +181,122 @@ def test_stop_timeout():
 
     """
 
-def test_start_test_program():
+
+def test_kill():
     """
-    >>> write('t.py',
-    ... '''
-    ... import time
-    ... time.sleep(1)
-    ... open('x', 'w').close()
-    ... time.sleep(99)
-    ... ''')
 
     >>> write('conf',
     ... '''
     ... <runner>
-    ...   program %s t.py
-    ...   start-test-program cat x
+    ...   program sleep 100
     ... </runner>
-    ... ''' % sys.executable)
-
-    >>> import os, time
-    >>> start = time.time()
+    ... ''')
 
     >>> system("./zdaemon -Cconf start")
     . .
-    daemon process started, pid=21446
+    daemon process started, pid=1234
 
-    >>> os.path.exists('x')
-    True
+    >>> system("./zdaemon -Cconf kill ded")
+    invalid signal 'ded'
+
+    >>> system("./zdaemon -Cconf kill CONT")
+    kill(1234, 18)
+    signal SIGCONT sent to process 1234
 
     >>> system("./zdaemon -Cconf stop")
-    <BLANKLINE>
+    . .
     daemon process stopped
+
+    >>> system("./zdaemon -Cconf kill")
+    daemon process not running
+
     """
 
+
+def test_logreopen():
+    """
+
+    >>> write('conf',
+    ... '''
+    ... <runner>
+    ...   program sleep 100
+    ...   transcript transcript.log
+    ... </runner>
+    ... ''')
+
+    >>> system("./zdaemon -Cconf start")
+    . .
+    daemon process started, pid=1234
+
+    >>> os.rename('transcript.log', 'transcript.log.1')
+
+    >>> system("./zdaemon -Cconf logreopen")
+    kill(1234, 12)
+    signal SIGUSR2 sent to process 1234
+
+    This also reopens the transcript.log:
+
+    >>> sorted(x
+    ...        for x in os.listdir('.')
+    ...        if x in [
+    ...            'conf', 'transcript.log', 'transcript.log.1', 'zdaemon', 'zdsock'])
+    ['conf', 'transcript.log', 'transcript.log.1', 'zdaemon', 'zdsock']
+
+    >>> system("./zdaemon -Cconf stop")
+    . .
+    daemon process stopped
+
+    """  # noqa: E501 line too long
+
+
+def test_log_rotation():
+    """
+
+    >>> write('conf',
+    ... '''
+    ... <runner>
+    ...   program sleep 100
+    ...   transcript transcript.log
+    ... </runner>
+    ... <eventlog>
+    ...   <logfile>
+    ...     path event.log
+    ...   </logfile>
+    ... </eventlog>
+    ... ''')
+
+    >>> system("./zdaemon -Cconf start")
+    . .
+    daemon process started, pid=1234
+
+    Pretend we did a logrotate:
+
+    >>> os.rename('transcript.log', 'transcript.log.1')
+    >>> os.rename('event.log', 'event.log.1')
+
+    >>> system("./zdaemon -Cconf reopen_transcript")  # or logreopen
+
+    This reopens both transcript.log and event.log:
+
+    >>> sorted(glob.glob('transcript.log*'))
+    ['transcript.log', 'transcript.log.1']
+
+    >>> sorted(glob.glob('event.log*'))
+    ['event.log', 'event.log.1']
+
+    >>> system("./zdaemon -Cconf stop")
+    . .
+    daemon process stopped
+
+    """
+
+
 def test_start_test_program():
     """
     >>> write('t.py',
     ... '''
     ... import time
-    ... time.sleep(1)
+    ... time.sleep(2)
     ... open('x', 'w').close()
     ... time.sleep(99)
     ... ''')
@@ -200,7 +306,15 @@ def test_start_test_program():
     ... <runner>
     ...   program %s t.py
     ...   start-test-program cat x
+    ...   daemon on
     ... </runner>
+    ... <eventlog>
+    ...   level debug
+    ...   <logfile>
+    ...      path log
+    ...      level debug
+    ...   </logfile>
+    ... </eventlog>
     ... ''' % sys.executable)
 
     >>> import os
@@ -213,17 +327,33 @@ def test_start_test_program():
     True
     >>> os.remove('x')
 
+    >>> with open("log") as f:
+    ...   logged = f.read()
+    >>> nfailed = logged.count("start test failed")
+    >>> nfailed > 0
+    True
+    >>> logged.count("start test succeeded")
+    1
+
     >>> system("./zdaemon -Cconf restart")
-    . . . 
+    . . .
     daemon process restarted, pid=19622
     >>> os.path.exists('x')
     True
 
+    >>> with open("log") as f:
+    ...   logged = f.read()
+    >>> nfailed < logged.count("start test failed")
+    True
+    >>> logged.count("start test succeeded")
+    2
+
     >>> system("./zdaemon -Cconf stop")
     <BLANKLINE>
     daemon process stopped
     """
 
+
 def test_start_timeout():
     """
     >>> write('t.py',
@@ -254,6 +384,7 @@ def test_start_timeout():
     daemon process stopped
     """
 
+
 def DAEMON_MANAGER_MODE_leak():
     """
     Zdaemon used an environment variable to flag that it's running in
@@ -274,6 +405,7 @@ def DAEMON_MANAGER_MODE_leak():
     True
     """
 
+
 def nonzero_exit_on_program_failure():
     """
     >>> write('conf',
@@ -317,15 +449,16 @@ def nonzero_exit_on_program_failure():
 
     """
 
+
 def setUp(test):
     test.globs['_td'] = td = []
     here = os.getcwd()
-    td.append(lambda : os.chdir(here))
+    td.append(lambda: os.chdir(here))
     tmpdir = tempfile.mkdtemp()
-    td.append(lambda : shutil.rmtree(tmpdir))
+    td.append(lambda: shutil.rmtree(tmpdir))
     test.globs['tmpdir'] = tmpdir
     workspace = tempfile.mkdtemp()
-    td.append(lambda : shutil.rmtree(workspace))
+    td.append(lambda: shutil.rmtree(workspace))
     os.chdir(workspace)
     write('zdaemon', zdaemon_template % dict(
         python=sys.executable,
@@ -335,6 +468,7 @@ def setUp(test):
     os.chmod('zdaemon', 0o755)
     test.globs['system'] = system
 
+
 def tearDown(test):
     for f in test.globs['_td']:
         f()
@@ -347,11 +481,13 @@ class Timeout(BaseException):
 @contextmanager
 def timeout(seconds):
     this_frame = sys._getframe()
+
     def raiseTimeout(signal, frame):
         # the if statement here is meant to prevent an exception in the
         # finally: clause before clean up can take place
         if frame is not this_frame:
             raise Timeout('timed out after %s seconds' % seconds)
+
     try:
         prev_handler = signal.signal(signal.SIGALRM, raiseTimeout)
     except ValueError:
@@ -385,9 +521,8 @@ def system(command, input='', quiet=False, echo=False):
 
 
 def checkenv(match):
-    match = [a for a in match.group(1).split('\n')[:-1]
-             if a.split('=')[0] in ('HOME', 'LD_LIBRARY_PATH')]
-    match.sort()
+    match = sorted([a for a in match.group(1).split('\n')[:-1]
+                    if a.split('=')[0] in ('HOME', 'LIBRARY_PATH')])
     return '\n'.join(match) + '\n'
 
 
@@ -399,37 +534,46 @@ sys.path[0:0] = [
   %(ZConfig)r,
   ]
 
+try:
+    import coverage
+except ModuleNotFoundError:
+    pass
+else:
+    coverage.process_startup()
+
 import zdaemon.zdctl
 
 if __name__ == '__main__':
     zdaemon.zdctl.main()
 """
 
+
 def test_suite():
     README_checker = renormalizing.RENormalizing([
-        (re.compile('pid=\d+'), 'pid=NNN'),
-        (re.compile('(\. )+\.?'), '<BLANKLINE>'),
+        (re.compile(r'pid=\d+'), 'pid=NNN'),
+        (re.compile(r'(\. )+\.?'), '<BLANKLINE>'),
         (re.compile('^env\n((?:.*\n)+)$'), checkenv),
-        ])
+    ])
 
     return unittest.TestSuite((
         doctest.DocTestSuite(
             setUp=setUp, tearDown=tearDown,
             checker=renormalizing.RENormalizing([
-                (re.compile('pid=\d+'), 'pid=NNN'),
-                (re.compile('(\. )+\.?'), '<BLANKLINE>'),
-                ])),
+                (re.compile(r'pid=\d+'), 'pid=NNN'),
+                (re.compile(r'(\. )+\.?'), '<BLANKLINE>'),
+                (re.compile(r'process \d+'), 'process NNN'),
+                (re.compile(r'kill\(\d+, \d+\)'), 'kill(NNN, MM)'),
+            ])),
         manuel.testing.TestSuite(
-                manuel.doctest.Manuel(
-                    parser=zc.customdoctests.DocTestParser(
-                        ps1='sh>',
-                        transform=lambda s: 'system("%s")\n' % s.rstrip()
-                        ),
-                    checker=README_checker,
-                    ) +
-                manuel.doctest.Manuel(checker=README_checker) +
-                manuel.capture.Manuel(),
-                '../README.rst',
-                setUp=setUp, tearDown=tearDown),
-        ))
-
+            manuel.doctest.Manuel(
+                parser=zc.customdoctests.DocTestParser(
+                    ps1='sh>',
+                    transform=lambda s: 'system("%s")\n' % s.rstrip()
+                ),
+                checker=README_checker,
+            ) +
+            manuel.doctest.Manuel(checker=README_checker) +
+            manuel.capture.Manuel(),
+            '../README.rst',
+            setUp=setUp, tearDown=tearDown),
+    ))
